@@ -3,23 +3,17 @@ try:
 except:
     import uheaqp as heapq
 
-try:
-    import time
-except:
-    import utime as time
+import gc
 
-PLACES_DICT = "places_dict"
-INV_TRANS_DICT = "inv_trans_dict"
-LABELS_DICT = "labels_dict"
-TRANS_LABELS_DICT = "trans_labels_dict"
-TRANS_PRE_DICT = "trans_pre_dict"
-TRANS_POST_DICT = "trans_post_dict"
-TRANSF_IM = "transf_im"
-TRANSF_FM = "transf_fm"
-TRANSF_MODEL_COST_FUNCTION = "transf_model_cost_function"
-TRANSF_TRACE = "transf_trace"
-TRACE_COST_FUNCTION = "trace_cost_function"
-INV_TRACE_LABELS_DICT = "inv_trace_labels_dict"
+TRANSF_MODEL_COST_FUNCTION = 0
+TRANS_PRE_DICT = 1
+TRANS_POST_DICT = 2
+LABELS_DICT = 3
+TRANS_LABELS_DICT = 4
+
+TRANSF_TRACE = 0
+TRACE_COST_FUNCTION = 1
+INV_TRACE_LABELS_DICT = 2
 
 IS_SYNC_MOVE = 0
 IS_LOG_MOVE = 1
@@ -55,13 +49,14 @@ def apply(trace, net, im, fm, model_cost_function=None, trace_cost_function=None
     ret_tuple_as_trans_desc
         Return tuple as trans desc
     """
-    model_struct = __transform_model_to_mem_efficient_structure(net, im, fm, model_cost_function=model_cost_function)
-    trace_struct = __transform_trace_to_mem_efficient_structure(trace, model_struct, trace_cost_function=trace_cost_function)
+    model_struct = __transform_model_to_mem_efficient_structure(net, model_cost_function=model_cost_function)
+    trace_struct = __transform_trace_to_mem_efficient_structure(trace, model_struct,
+                                                                trace_cost_function=trace_cost_function)
 
-    return __dijkstra(model_struct, trace_struct, net, ret_tuple_as_trans_desc=ret_tuple_as_trans_desc)
+    return __dijkstra(model_struct, trace_struct, net, im, fm, ret_tuple_as_trans_desc=ret_tuple_as_trans_desc)
 
 
-def __transform_model_to_mem_efficient_structure(net, im, fm, model_cost_function=None):
+def __transform_model_to_mem_efficient_structure(net, model_cost_function=None):
     """
     Transform the Petri net model to a memory efficient structure
 
@@ -69,10 +64,6 @@ def __transform_model_to_mem_efficient_structure(net, im, fm, model_cost_functio
     --------------
     net
         Petri net
-    im
-        Initial marking
-    fm
-        Final marking
     model_cost_function
         Model cost function
 
@@ -80,22 +71,19 @@ def __transform_model_to_mem_efficient_structure(net, im, fm, model_cost_functio
     --------------
     model_struct
         Model data structure, including:
-            PLACES_DICT: associates each place to a number
-            INV_TRANS_DICT: associates a number to each transition
-            LABELS_DICT: labels dictionary (a label to a number)
-            TRANS_LABELS_DICT: associates each transition to the number corresponding to its label
+            TRANSF_MODEL_COST_FUNCTION: transformed model cost function
             TRANS_PRE_DICT: preset of a transition, expressed as in this data structure
             TRANS_POST_DICT: postset of a transition, expressed as in this data structure
-            TRANSF_IM: transformed initial marking
-            TRANSF_FM: transformed final marking
-            TRANSF_MODEL_COST_FUNCTION: transformed model cost function
+            LABELS_DICT: labels dictionary (a label to a number)
+            TRANS_LABELS_DICT: associates each transition to the number corresponding to its label
     """
     if model_cost_function is None:
-        model_cost_function = {}
+        model_cost_function = []
         i = 0
         while i < len(net[1]):
-            model_cost_function[i] = 10000
+            model_cost_function.append(10000)
             i = i + 1
+        model_cost_function = tuple(model_cost_function)
 
     trans_pre_dict = {}
     trans_post_dict = {}
@@ -108,15 +96,14 @@ def __transform_model_to_mem_efficient_structure(net, im, fm, model_cost_functio
         i = i + 1
     labels = sorted(list(labels))
     labels_dict = {labels[i]: i for i in range(len(labels))}
-    trans_labels_dict = {}
+    trans_labels_dict = []
     i = 0
     while i < len(net[1]):
-        trans_labels_dict[i] = labels_dict[net[1][i][0]]
+        trans_labels_dict.append(labels_dict[net[1][i][0]])
         i = i + 1
+    trans_labels_dict = tuple(trans_labels_dict)
 
-    return {TRANSF_MODEL_COST_FUNCTION: model_cost_function, TRANS_PRE_DICT: trans_pre_dict,
-            TRANS_POST_DICT: trans_post_dict, TRANSF_IM: im,
-            TRANSF_FM: fm, LABELS_DICT: labels_dict, TRANS_LABELS_DICT: trans_labels_dict}
+    return (model_cost_function, trans_pre_dict, trans_post_dict, labels_dict, trans_labels_dict)
 
 
 def __transform_trace_to_mem_efficient_structure(trace, model_struct, trace_cost_function=None):
@@ -153,8 +140,8 @@ def __transform_trace_to_mem_efficient_structure(trace, model_struct, trace_cost
     transf_trace = [labels_dict[x] for x in trace[1]]
 
     inv_trace_labels_dict = {y: x for x, y in labels_dict.items()}
-    return {TRANSF_TRACE: transf_trace, TRACE_COST_FUNCTION: trace_cost_function,
-            INV_TRACE_LABELS_DICT: inv_trace_labels_dict}
+
+    return (transf_trace, trace_cost_function, inv_trace_labels_dict)
 
 
 def __dict_leq(d1, d2):
@@ -215,14 +202,14 @@ def __fire_trans(m, preset, postset):
     return ret
 
 
-def __encode_marking(marking_dict, m_d):
+def __encode_marking(existing_markings, m_d):
     """
     Encode a marking using the dictionary
 
     Parameters
     --------------
-    marking_dict
-        Marking dictionary
+    existing_markings
+        Existing markings
     m_d
         Current marking (dict)
 
@@ -237,9 +224,13 @@ def __encode_marking(marking_dict, m_d):
         for i in range(m_d[el]):
             m_t.append(el)
     m_t = tuple(m_t)
-    if m_t not in marking_dict:
-        marking_dict[m_t] = m_t
-    return marking_dict[m_t]
+    i = 0
+    while i < len(existing_markings):
+        if existing_markings[i] == m_t:
+            return existing_markings, existing_markings[i]
+        i = i + 1
+    existing_markings = existing_markings + (m_t,)
+    return existing_markings, m_t
 
 
 def __decode_marking(m_t):
@@ -266,8 +257,65 @@ def __decode_marking(m_t):
     return m_d
 
 
-def __dijkstra(model_struct, trace_struct, net, sync_cost=0, max_align_time_trace=10000000,
-               ret_tuple_as_trans_desc=False):
+def __check_closed(closed, ns):
+    """
+    Checks if the state is closed
+
+    Parameters
+    -------------
+    closed
+        Closed set
+    ns
+        New state (marking, index)
+
+    Returns
+    -------------
+    bool
+        Boolean (true if the state is closed)
+    """
+    i = 0
+    while i < len(closed):
+        if closed[i][0] == ns[0] and closed[i][1] <= ns[1]:
+            return True
+        i = i + 1
+    return False
+
+
+def __add_to_open_set(open_set, ns):
+    """
+    Adds a new state to the open set whether necessary
+
+    Parameters
+    ----------------
+    open_set
+        Open set
+    ns
+        New state
+    """
+    shall_add = True
+    shall_heapify = False
+    i = 0
+    while i < len(open_set):
+        if open_set[i][POSITION_MARKING] == ns[POSITION_MARKING]:
+            if open_set[i][POSITION_INDEX] <= ns[POSITION_INDEX] and open_set[i][POSITION_TOTAL_COST] <= ns[
+                POSITION_TOTAL_COST]:
+                # do not add anything
+                shall_add = False
+                break
+            if open_set[i][POSITION_INDEX] >= ns[POSITION_INDEX] and open_set[i][POSITION_TOTAL_COST] > ns[
+                POSITION_TOTAL_COST]:
+                del open_set[i]
+                shall_heapify = True
+                continue
+        i = i + 1
+    if shall_add:
+        heapq.heappush(open_set, ns)
+    if shall_heapify:
+        heapq.heapify(open_set)
+    return open_set
+
+
+def __dijkstra(model_struct, trace_struct, net, im, fm, sync_cost=0, ret_tuple_as_trans_desc=False, enable_gc=True):
     """
     Alignments using Dijkstra
 
@@ -279,13 +327,17 @@ def __dijkstra(model_struct, trace_struct, net, sync_cost=0, max_align_time_trac
         Efficient trace structure
     net
         Petri net
+    im
+        Initial marking
+    fm
+        Final marking
     sync_cost
         Cost of a sync move (limitation: all sync moves shall have the same cost in this setting)
-    max_align_time_trace
-        Maximum alignment time for a trace (in seconds)
     ret_tuple_as_trans_desc
         Says if the alignments shall be constructed including also
         the name of the transition, or only the label (default=False includes only the label)
+    enable_gc
+        Boolean value (enabling or not the garbage collection)
 
     Returns
     --------------
@@ -296,8 +348,6 @@ def __dijkstra(model_struct, trace_struct, net, sync_cost=0, max_align_time_trac
             visited: the number of states that have been visited
             cost: the cost of the alignment
     """
-    start_time = time.time()
-
     trans_pre_dict = model_struct[TRANS_PRE_DICT]
     trans_post_dict = model_struct[TRANS_POST_DICT]
     trans_labels_dict = model_struct[TRANS_LABELS_DICT]
@@ -306,11 +356,11 @@ def __dijkstra(model_struct, trace_struct, net, sync_cost=0, max_align_time_trac
     transf_trace = trace_struct[TRANSF_TRACE]
     trace_cost_function = trace_struct[TRACE_COST_FUNCTION]
 
-    marking_dict = {}
-    closed = set()
+    existing_markings = tuple([])
+    closed = tuple([])
 
-    im = __encode_marking(marking_dict, model_struct[TRANSF_IM])
-    fm = __encode_marking(marking_dict, model_struct[TRANSF_FM])
+    existing_markings, im = __encode_marking(existing_markings, im)
+    existing_markings, fm = __encode_marking(existing_markings, fm)
 
     # each state is characterized by:
     # position 0 (POSITION_TOTAL_COST): total cost of the state
@@ -332,57 +382,65 @@ def __dijkstra(model_struct, trace_struct, net, sync_cost=0, max_align_time_trac
     visited = 0
 
     while not len(open_set) == 0:
-        if (time.time() - start_time) > max_align_time_trace:
-            return None
         curr = heapq.heappop(open_set)
-        curr_m0 = curr[POSITION_MARKING]
-        if (curr_m0, curr[POSITION_INDEX]) in closed:
-            continue
-        closed.add((curr_m0, curr[POSITION_INDEX]))
-        curr_m = __decode_marking(curr_m0)
         visited = visited + 1
-        if curr_m0 == fm:
+        if __check_closed(closed, (curr[POSITION_MARKING], curr[POSITION_INDEX])):
+            continue
+        if curr[POSITION_MARKING] == fm:
             if -curr[POSITION_INDEX] == len(transf_trace):
                 # returns the alignment only if the final marking has been reached AND
                 # the trace is over
-                return __reconstruct_alignment(curr, model_struct, trace_struct, visited, net,
+                return __reconstruct_alignment(curr, trace_struct, net, visited, len(open_set), len(closed),
+                                               len(existing_markings),
                                                ret_tuple_as_trans_desc=ret_tuple_as_trans_desc)
+            else:
+                closed = closed + ((curr[POSITION_MARKING], curr[POSITION_INDEX]),)
         else:
+            closed = closed + ((curr[POSITION_MARKING], curr[POSITION_INDEX]),)
+            curr_m = __decode_marking(curr[POSITION_MARKING])
             # retrieves the transitions that are enabled in the current marking
-            en_t = [t for t in trans_pre_dict if __dict_leq(trans_pre_dict[t], curr_m)]
+            en_t = tuple([t for t in trans_pre_dict if __dict_leq(trans_pre_dict[t], curr_m)])
             for t in en_t:
                 # checks if a given transition can be executed in sync with the trace
                 is_sync = trans_labels_dict[t] == transf_trace[-curr[POSITION_INDEX]] if -curr[POSITION_INDEX] < len(
                     transf_trace) else False
                 # virtually fires the transition to get a new marking
-                new_m = __encode_marking(marking_dict, __fire_trans(curr_m, trans_pre_dict[t], trans_post_dict[t]))
+                existing_markings, new_m = __encode_marking(existing_markings,
+                                                            __fire_trans(curr_m, trans_pre_dict[t], trans_post_dict[t]))
                 if is_sync:
                     dummy_count = dummy_count + 1
-                    if (new_m, curr[POSITION_INDEX]-1) not in closed:
+                    if not __check_closed(closed, (new_m, curr[POSITION_INDEX] - 1)):
                         new_state = (
                             curr[POSITION_TOTAL_COST] + sync_cost, curr[POSITION_INDEX] - 1, IS_SYNC_MOVE, dummy_count,
                             curr,
                             new_m, t)
-                        heapq.heappush(open_set, new_state)
-                dummy_count = dummy_count + 1
-                new_state = (
-                    curr[POSITION_TOTAL_COST] + transf_model_cost_function[t], curr[POSITION_INDEX], IS_MODEL_MOVE,
-                    dummy_count, curr, new_m, t)
-                heapq.heappush(open_set, new_state)
+                        open_set = __add_to_open_set(open_set, new_state)
+                else:
+                    # avoid scheduling a move-on-model when model and trace are sync.
+                    dummy_count = dummy_count + 1
+                    if not __check_closed(closed, (new_m, curr[POSITION_INDEX] - 1)):
+                        new_state = (
+                            curr[POSITION_TOTAL_COST] + transf_model_cost_function[t], curr[POSITION_INDEX],
+                            IS_MODEL_MOVE,
+                            dummy_count, curr, new_m, t)
+                        open_set = __add_to_open_set(open_set, new_state)
         # IMPORTANT: to reduce the complexity, assume that you can schedule a log move
         # only if the previous move has not been a move-on-model.
         # since this setting is equivalent to scheduling all the log moves before and then
         # the model moves
         if -curr[POSITION_INDEX] < len(transf_trace) and curr[POSITION_TYPE_MOVE] != IS_MODEL_MOVE:
             dummy_count = dummy_count + 1
-            if (curr_m0, curr[POSITION_INDEX] - 1) not in closed:
+            if not __check_closed(closed, (curr[POSITION_MARKING], curr[POSITION_INDEX] - 1)):
                 new_state = (
                     curr[POSITION_TOTAL_COST] + trace_cost_function[-curr[POSITION_INDEX]], curr[POSITION_INDEX] - 1,
-                    IS_LOG_MOVE, dummy_count, curr, curr_m0, None)
-                heapq.heappush(open_set, new_state)
+                    IS_LOG_MOVE, dummy_count, curr, curr[POSITION_MARKING], None)
+                open_set = __add_to_open_set(open_set, new_state)
+        if enable_gc:
+            gc.collect()
 
 
-def __reconstruct_alignment(curr, model_struct, trace_struct, visited, net, ret_tuple_as_trans_desc=False):
+def __reconstruct_alignment(curr, trace_struct, net, visited, queued, closed_set_length, num_visited_markings,
+                            ret_tuple_as_trans_desc=False):
     """
     Reconstruct the alignment from the final state (that reached the final marking)
 
@@ -390,14 +448,18 @@ def __reconstruct_alignment(curr, model_struct, trace_struct, visited, net, ret_
     ----------------
     curr
         Current state (final state)
-    model_struct
-        Efficient data structure for the model
     trace_struct
         Efficient data structure for the trace
-    visited
-        Number of visited states
     net
         Petri net
+    visited
+        Number of visited states
+    queued
+        Number of queued states (at the end of execution; shall be summed by visited)
+    closed_set_length
+        Length of the closed set
+    num_visited_markings
+        Number of visited markings
     ret_tuple_as_trans_desc
         Says if the alignments shall be constructed including also
         the name of the transition, or only the label (default=False includes only the label)
@@ -416,7 +478,8 @@ def __reconstruct_alignment(curr, model_struct, trace_struct, visited, net, ret_
 
     alignment = []
     cost = curr[POSITION_TOTAL_COST]
-    queued = curr[POSITION_STATES_COUNT]
+
+    queued = queued + visited
 
     while curr[POSITION_PARENT_STATE] is not None:
         m_name, m_label, t_name, t_label = ">>", ">>", ">>", ">>"
@@ -433,4 +496,5 @@ def __reconstruct_alignment(curr, model_struct, trace_struct, visited, net, ret_
             alignment = [(t_label, m_label)] + alignment
         curr = curr[POSITION_PARENT_STATE]
 
-    return {"alignment": alignment, "cost": cost, "queued_states": queued, "visited_states": visited}
+    return {"alignment": alignment, "cost": cost, "queued_states": queued, "visited_states": visited,
+            "closed_set_length": closed_set_length, "num_visited_markings": num_visited_markings}
